#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h> 

#include <errno.h>

#include "_cdio_stream.h"
#include "_cdio_stdio.h"
#include "cdio/util.h"

#include <bm.h>

typedef struct
{
	block_t *blk;
	off_t offset;
	uint64_t st_size;
	int	align;
} cd_blk_t;

static int blk_io_open (void *user_data)
{
	cd_blk_t *const ud = user_data;
	return (ud->blk == NULL);
}

static int blk_io_close(void *user_data)
{
	return 0;
}

static void blk_io_free(void *user_data)
{
	cd_blk_t *const ud = user_data;
	ud->blk = NULL;
	free(ud);
}

/*! 
  Like fseek(3) and in fact may be the same.
  
  This  function sets the file position indicator for the stream
  pointed to by stream.  The new position, measured in bytes, is obtained
  by  adding offset bytes to the position specified by whence.  If whence
  is set to SEEK_SET, SEEK_CUR, or SEEK_END, the offset  is  relative  to
  the  start of the file, the current position indicator, or end-of-file,
  respectively.  A successful call to the fseek function clears the end-
  of-file indicator for the stream and undoes any effects of the
  ungetc(3) function on the same stream.
  
  @return upon successful completion, DRIVER_OP_SUCCESS, else,
  DRIVER_OP_ERROR is returned and the global variable errno is set to
  indicate the error.
*/
static driver_return_code_t 
blk_io_seek(void *p_user_data, long i_offset, int whence)
{
	cd_blk_t *const ud = p_user_data;

	if (whence == SEEK_CUR)
		ud->offset += i_offset;
	else if (whence == SEEK_SET)
		ud->offset = i_offset;
	else if (whence == SEEK_END)
		ud->offset = ud->st_size + i_offset;

	return DRIVER_OP_SUCCESS;
}

static long int blk_io_stat(void *p_user_data)
{
  const cd_blk_t *const ud = p_user_data;

  return ud->st_size;
}

#define IS_ALIGNED(n, align) ((((uintptr_t) n) & \
			       ((uintptr_t) ((align)-1))) == 0)

/*!
  Like fread(3) and in fact is about the same.
  
  DESCRIPTION:
  The function fread reads nmemb elements of data, each size bytes long,
  from the stream pointed to by stream, storing them at the location
  given by ptr.
  
  RETURN VALUE:
  return the number of items successfully read or written (i.e.,
  not the number of characters).  If an error occurs, or the
  end-of-file is reached, the return value is a short item count
  (or zero).
  
  We do not distinguish between end-of-file and error, and callers
  must use feof(3) and ferror(3) to determine which occurred.
  */
static long blk_io_read(void *user_data, void *buf, long int size)
{
	cd_blk_t *const ud = user_data;
	ssize_t ret, actual = 0;

	if (ud->align && (!IS_ALIGNED(ud->offset, ud->align) || IS_ALIGNED(size, ud->align)))
	{
		/*
		 * The location which we're trying to read isn't aligned,
		 * so we need to do this the hard way...
		 */

		char tmp[ud->align];

		do
		{
			size_t align_offset = ud->offset % ud->align;
			ret = ud->blk->pread(ud->blk, tmp, ud->align, ud->offset-align_offset);
			if (ret <= 0)
				return -EIO;

			int need = MIN(ud->align - align_offset, size);
			memmove(buf, tmp+align_offset, need);
			buf += need;
			size -= need;
			ud->offset += need;
			actual += need;
		} while (size > 0 && !IS_ALIGNED(size, ud->align));
	}

	if (size > 0)
	{
		ret = ud->blk->pread(ud->blk, buf, size, ud->offset);
		if (ret <= 0)
			return -EIO;

		actual += ret;
		ud->offset += ret; 
	}

	return actual;
}

/*!
  Deallocate resources assocaited with obj. After this obj is unusable.
*/
void cdio_stdio_destroy(CdioDataSource_t *p_obj)
{
	cdio_stream_destroy(p_obj);
}

CdioDataSource_t* cdio_stdio_new(const char *pathname)
{
	cdio_stream_io_functions funcs = { NULL, NULL, NULL, NULL, NULL, NULL };
	cd_blk_t *ud = malloc(sizeof(cd_blk_t));

	ud->blk = (block_t *)pathname;
	ud->blk->ioctl(ud->blk, IOCTL_GET_SIZE, &ud->st_size);

	uint32_t sector_size = 4096;
	ud->blk->ioctl(ud->blk, IOCTL_GET_SECTOR_SIZE, &sector_size);
	ud->align = sector_size;
	ud->offset = 0;

	funcs.open   = blk_io_open;
	funcs.seek   = blk_io_seek;
	funcs.stat   = blk_io_stat;
	funcs.read   = blk_io_read;
	funcs.close  = blk_io_close;
	funcs.free   = blk_io_free;

	return cdio_stream_new(ud, &funcs);
}
