/*
 * CD FS support
 * 
 * Copyright (c) 2021-2022 Ant-ON
 *
 */

#include <sys/param.h>
#include <string.h>
#include <stdlib.h>
#include <fcntl.h>
#include <errno.h>

#include <fs.h>

#include <cdio/iso9660.h>
#include <cdio/udf.h>
#include <cdio/udf_file.h>
#include "lib/udf/udf_private.h"

typedef struct
{
	udf_t *udf;
	udf_dirent_t *udf_droot;
	iso9660_t *iso;
	uint64_t size;
} cd_fs_t;

typedef struct
{
	cd_fs_t *fs;
	
	udf_dirent_t *udf_dirent;

	lsn_t lsn;
	uint64_t size;              /* Total length of the file (in bytes) */
	off_t offset;               /* Current position within the file (in bytes) */
} cd_file_t;

static inline ssize_t cd_f_read_blk(cd_file_t *cd_fd, void *buf, size_t len, off_t offset)
{
	if (cd_fd->udf_dirent)
	{
		cd_fd->udf_dirent->p_udf->i_position = offset;
		return udf_read_block(cd_fd->udf_dirent, buf, len/UDF_BLOCKSIZE);
	} else
	{
		return iso9660_iso_seek_read(cd_fd->fs->iso, buf, cd_fd->lsn+(offset/ISO_BLOCKSIZE), len/ISO_BLOCKSIZE);
	}
}

static ssize_t cd_f_read(void *vfd, void *buf, size_t len)
{
	fd_t *fd = (fd_t *)vfd;
	cd_file_t *cd_fd = fd->data;

	size_t size = len;
	ssize_t actual;

#if ISO_BLOCKSIZE != UDF_BLOCKSIZE
#error "Unsupported!"
#endif

	size_t offset = cd_fd->offset % ISO_BLOCKSIZE;
	if (offset)
	{
		char tmp[ISO_BLOCKSIZE];

		off_t aligned_off = cd_fd->offset - offset;
		int need = MIN(ISO_BLOCKSIZE - offset, size);

		actual = cd_f_read_blk(cd_fd, tmp, ISO_BLOCKSIZE, aligned_off);
		if (actual <= 0)
			return -EIO;
		
		memcpy(buf, tmp+offset, need);
		buf += need;
		size -= need;
		cd_fd->offset += need;
	}

	if (size > 0)
	{
		size_t unaligned_size = size % ISO_BLOCKSIZE;
		size_t aligned_size = size - unaligned_size;
		
		actual = cd_f_read_blk(cd_fd, buf, aligned_size, cd_fd->offset);
		if (actual <= 0)
			return -EIO;

		cd_fd->offset += aligned_size;

		if (unaligned_size)
		{
			char tmp[ISO_BLOCKSIZE];
			
			actual = cd_f_read_blk(cd_fd, tmp, ISO_BLOCKSIZE, cd_fd->offset);
			if (actual <= 0)
				return -EIO;

			memcpy(buf + aligned_size, tmp, unaligned_size);
			cd_fd->offset += unaligned_size;
		}
	}

	return len;
}

static off_t cd_f_seek(void *vfd, off_t offset, int whence)
{
	fd_t *fd = (fd_t *)vfd;
	cd_file_t *cd_fd = fd->data;

	if (whence == SEEK_CUR)
		cd_fd->offset += offset;
	else if (whence == SEEK_SET)
		cd_fd->offset = offset;
	else if (whence == SEEK_END)
		cd_fd->offset = cd_fd->size + offset;

    return cd_fd->offset;
}

static void cd_f_close(void *vfd)
{
	fd_t *fd = (fd_t *)vfd;
	cd_file_t *cd_fd = fd->data;
	if (!cd_fd)
		return;
	fd->data = NULL;

	if (cd_fd->udf_dirent)
		udf_dirent_free(cd_fd->udf_dirent);
	free(cd_fd);
	//free(fd); - free in FS.c
}

static void* cd_fs_fopen(void *vfs, const char *path, int flags, long *size)
{
	fs_t *fs = (fs_t *)vfs;
	cd_fs_t *cd_fs = fs->data;

	if ((flags&O_ACCMODE) != O_RDONLY)
		return NULL;

	cd_file_t *cd_fd = (cd_file_t *)malloc(sizeof(cd_file_t));

	if (cd_fs->udf)
	{
		cd_fd->lsn = 0;
		cd_fd->size = 0;

		//udf_dirent_t *udf_dirent_root = udf_get_root(cd_fs->udf, true, 0);
		//cd_fd->udf_dirent = udf_fopen(udf_dirent_root, path);
		cd_fd->udf_dirent = udf_fopen(cd_fs->udf_droot, path);
		if (!cd_fd->udf_dirent)
			goto error;
		cd_fd->size = udf_get_file_length(cd_fd->udf_dirent);
	} else
	{
		cd_fd->udf_dirent = NULL;

		iso9660_stat_t *stat = iso9660_ifs_stat(cd_fs->iso, path);
		if (!stat)
			goto error;
		
		cd_fd->lsn = stat->lsn;
		cd_fd->size = stat->size;

		iso9660_stat_free(stat);
	}
	cd_fd->fs = cd_fs;
	cd_fd->offset = 0;

	fd_t *fd = malloc(sizeof(fd_t));
	fd->read = cd_f_read;
	fd->write = NULL;
	fd->seek = cd_f_seek;
	fd->close = cd_f_close;
	fd->data = cd_fd;
	if (size)
		*size = cd_fd->size;
	return fd;

error:
	if (cd_fd)
		free(cd_fd);
	return NULL;
}

static int cd_fs_list(void *vfs, void *ctx, const char *path, _filter filter)
{
	fs_t *fs = (fs_t *)vfs;
	cd_fs_t *cd_fs = fs->data;

	if (cd_fs->udf)
	{
		udf_dirent_t *udf_dirent, *udf_dirent_path = NULL;
		if (strcmp("/", path))
		{
			udf_dirent_path = udf_fopen(cd_fs->udf_droot, path);
			if (!udf_dirent_path)
				return -ENOENT;
			udf_dirent = udf_opendir(udf_dirent_path);
			if (!udf_dirent)
				return -ENOENT;
		} else
			udf_dirent = udf_get_root(cd_fs->udf, true, 0);

		while (udf_readdir(udf_dirent)) 
		{
			const char *name = udf_get_filename(udf_dirent);
			if (!name || !name[0])
				continue;

			filter(ctx, udf_get_filename(udf_dirent), 
				udf_get_file_length(udf_dirent), 
				udf_get_modification_time(udf_dirent), 
				udf_is_dir(udf_dirent)?F_FLAG_DIR:0);
		}
		
		if (udf_dirent_path)
			udf_dirent_free(udf_dirent_path);
	} else
	{
		CdioList_t *list = iso9660_ifs_readdir(cd_fs->iso, path);
		if (!list)
			return -ENOENT;

		CdioListNode_t *node;
		_CDIO_LIST_FOREACH (node, list)
		{
			char filename[4096];
			iso9660_stat_t *stat = (iso9660_stat_t *) _cdio_list_node_data (node);
			iso9660_name_translate(stat->filename, filename);

			if (filename[0] == '\0' || (filename[0] == '.' && (filename[1] == '\0' || filename[1] == '.')))
				continue;

			int flags = (stat->type == _STAT_DIR)?F_FLAG_DIR:0;
			filter(ctx, filename, stat->size, mktime(&stat->tm), flags);
		}

		_cdio_list_free (list, true);
	}

	return 0;
}

static int cd_fs_stat(void *vfs, const char *path, stat_t *stat)
{
	fs_t *fs = (fs_t *)vfs;
	cd_fs_t *cd_fs = fs->data;

	if (cd_fs->udf)
	{
		udf_dirent_t *udf_dirent = udf_fopen(cd_fs->udf_droot, path);
		if (!udf_dirent)
			return -ENOENT;
		
		stat->size = udf_get_file_length(udf_dirent);
		stat->time = udf_get_modification_time(udf_dirent);
		stat->flags = udf_is_dir(udf_dirent)?F_FLAG_DIR:0;

		udf_dirent_free(udf_dirent);
	} else
	{
		iso9660_stat_t *st = iso9660_ifs_stat(cd_fs->iso, path);
		if (!st)
			return -ENOENT;

		stat->size = st->size;
		stat->time = mktime(&st->tm);
		stat->flags = (st->type == _STAT_DIR)?F_FLAG_DIR:0;

		iso9660_stat_free(st);
	}

	return 0;
}

static int cd_fs_statfs(void *vfs, stat_fs_t *stat_fs)
{
	fs_t *fs = (fs_t *)vfs;
	cd_fs_t *cd_fs = fs->data;

	stat_fs->name = NULL;
	stat_fs->size = cd_fs->size;
	stat_fs->free = 0;
	stat_fs->flags = FLAG_FS_OPEN_RO;
	return 0;
}

static void cd_fs_fsfree(void *vfs)
{
	fs_t *fs = (fs_t *)vfs;
	cd_fs_t *cd_fs = fs->data;

	if (cd_fs->udf)
	{
		udf_close(cd_fs->udf);
		if (cd_fs->udf_droot)
			udf_dirent_free(cd_fs->udf_droot);
	} else
	{
		iso9660_close(cd_fs->iso);
	}

	free(fs->data);
	free(fs);
}

#ifdef MODULE_FS
void* __UI fs_open(void *block, int flags)
#else
void* __UI cd_fs_open(void *block, int flags)
#endif
{
	cd_fs_t *cd_fs;
	iso9660_t *p_iso = NULL;
	udf_dirent_t *udf_droot = NULL;
	block_t *blk = (block_t *)block;
	udf_t *p_udf = udf_open(block);
	if (p_udf)
	{
		udf_droot = udf_get_root(p_udf, true, 0);
		if (!udf_droot)
		{
			udf_close(p_udf);
			return NULL;
		}
	} else
	{
		p_iso = iso9660_open_ext(block, ISO_EXTENSION_JOLIET);
		if (!p_iso)
			return NULL;
	}

	cd_fs = (cd_fs_t *)malloc(sizeof(cd_fs_t));
	if (!cd_fs)
		return NULL;

	blk->ioctl(blk, IOCTL_GET_SIZE, &cd_fs->size);
	
	cd_fs->iso = p_iso;
	cd_fs->udf = p_udf;
	cd_fs->udf_droot = udf_droot;

	fs_t *fs = malloc(sizeof(fs_t));
	fs->open = cd_fs_fopen;
	fs->mkdir = NULL;
	fs->rename = NULL;
	fs->remove = NULL;
	fs->list = cd_fs_list;
	fs->stat = cd_fs_stat;
	fs->statfs = cd_fs_statfs;
	fs->list = cd_fs_list;
	fs->fsfree = cd_fs_fsfree;
	fs->data = cd_fs;
	return fs;
}
